/* 
 * Sample program for use with Product          
 *  ProgIds: 5724-J06 5724-J05 5724-J04 5697-J09 5655-M74 5655-M75 5648-C63 
 *  (C) Copyright IBM Corporation 2005.                      
 * All Rights Reserved * Licensed Materials - Property of IBM 
 * 
 * This sample program is provided AS IS and may be used, executed, 
 * copied and modified without royalty payment by customer 
 * 
 * (a) for its own instruction and study, 
 * (b) in order to develop applications designed to run with an IBM 
 *     WebSphere product, either for customer's own internal use or for 
 *     redistribution by customer, as part of such an application, in 
 *     customer's own products. 
 */ 
 
package com.ibm.broker.javacompute.samples;
import java.util.ListResourceBundle;

import com.google.soap.search.*;
import com.ibm.broker.javacompute.MbJavaComputeNode;
import com.ibm.broker.plugin.*;

/**
 * 
 * The GoogleAPINode sample demonstrates how a Java compute node can call out to
 * an external service and propagate a new message based on the results of this call.
 * 
 * <p>
 * <b>N.B. The example uses the Google API and requires googleapi.jar and a
 * key that can be obtained by registering at http://www.google.com/apis/
 * 
 * <p>
 * The incoming messages is of the following format:
 * 
 * <googleapinode>
 *   <key> </key> <!-- This is the key obtained by registering -->
 *   <querystring>test</querystring> <!-- The string to search for --> 
 * </googleapinode>
 * 
 * <p>
 * The node extracts the "key" and "querystring" element values from the incoming message. It uses this
 * information to perform a Google search. The result of this search is used to
 * create the outgoing message of the following form:
 * 
 * <googlesearchresult>
 *   <resultelement>
 *       <url>...</url>
 *       <title>...</title>
 *       <snippet>...</snippet>
 *       <summary>...</summary>
 *   </resultelement>
 *   ...
 * </googlesearchresult>
 * 
 */
public class GoogleAPINode extends MbJavaComputeNode
{
  /* (non-Javadoc)
   * @see com.ibm.broker.javacompute.MbJavaComputeNode#evaluate(com.ibm.broker.plugin.MbMessageAssembly)
   */
  public void evaluate(MbMessageAssembly incomingAssembly) throws MbException
  {
    final String methodName = "evaluate";
    
    // First use XPath expressions to extract the "querystring" and "key" element values
    // from the incoming message.
    // The format of the incoming message is as follows:
    // <googleapinode>
    //   <key> </key> <!-- This is the key obtained by registering -->
    //   <querystring>test</querystring> <!-- The string to search for --> 
    // </googleapinode>
    //
    String queryString = (String)incomingAssembly.getMessage().evaluateXPath("string(/googleapinode/querystring)");
    String keyString = (String)incomingAssembly.getMessage().evaluateXPath("string(/googleapinode/key)");
    
    try
    {
      // Perform the Google search
      GoogleSearchResult result = performSearch(keyString, queryString);
      
      // Build the output message based on the search result.
      MbMessage searchResultMessage = buildSearchResultMessage(incomingAssembly.getMessage(), result);
      
      // Propagate the augmented message. 
      getOutputTerminal("out").propagate(new MbMessageAssembly(incomingAssembly, searchResultMessage));
      
    }
    catch (GoogleSearchFault e)
    {
      e.printStackTrace();
      // An exception has been thrown by the Google API so log the exception and stacktrace
      // and throw a MbUserException so the message is propagated to the failure terminal.
      String stackTrace = stackTraceToString(e);
      logErrorGenerateException(this, 
                                methodName, 
                                GoogleAPINodeMesssages.MESSAGE_SOURCE, 
                                GoogleAPINodeMesssages.GOOGLE_API_EXCEPTION, 
                                "Google API exception", 
                                new String[] { keyString, queryString, e.toString(), stackTrace });
    }
  }
  
  public void copyMessageHeaders(MbMessage inMessage, MbMessage outMessage) throws MbException
  {
    MbElement outRoot = outMessage.getRootElement();
    MbElement header = inMessage.getRootElement().getFirstChild();

    while(header != null && header.getNextSibling() != null)
      {
        outRoot.addAsLastChild(header.copy());
        header = header.getNextSibling();
      }
  }
  
  /**
   * Creates a new result message based on the Google search result.
   * 
   * @param incomingMessage The incoming message.
   * @param result The Google search result.
   * @return The new result message.
   * @throws MbException
   */
  private MbMessage buildSearchResultMessage(MbMessage incomingMessage,
                                             GoogleSearchResult result) throws MbException
  {
    // Create an new empty message.
    MbMessage searchResultMessage = new MbMessage();
    // Copy headers from the input message
    copyMessageHeaders(incomingMessage, searchResultMessage); 
    MbElement outRoot = searchResultMessage.getRootElement();
    // create the "XML" element, and its child "googlesearchresult"
    MbElement googleSearchResultElement = outRoot.createElementAsLastChild("XML").createElementAsLastChild(MbElement.TYPE_NAME, "googlesearchresult", null);
    
    GoogleSearchResultElement[] results = result.getResultElements();
    
    // Iterate over the GoogleSearchResultElement objects.
    for (int i = 0; i < results.length; i++)
    {
      MbElement resultElement = googleSearchResultElement.createElementAsLastChild(MbElement.TYPE_NAME, "resultelement", null);
      GoogleSearchResultElement element = results[i];
      
      // Create elements in the outgoing message
      resultElement.createElementAsLastChild(MbElement.TYPE_NAME, "url", element.getURL());
      resultElement.createElementAsLastChild(MbElement.TYPE_NAME, "title", element.getTitle());
      resultElement.createElementAsLastChild(MbElement.TYPE_NAME, "snippet", element.getSnippet());
      resultElement.createElementAsLastChild(MbElement.TYPE_NAME, "summary", element.getSummary());
    }
    return searchResultMessage;
  }

  
  /**
   * Calls out to the Google API to perform the search.
   *  
   * @param key The client key required by the Goggle API ( see http://www.Google.com/apis/ ).
   * @param queryString The search string.
   * @return The Google search result.
   * @throws GoogleSearchFault
   */
  private static GoogleSearchResult performSearch(String key, 
                                    String queryString) throws GoogleSearchFault
  {
    GoogleSearch s = new GoogleSearch();
    s.setKey(key);
    s.setQueryString(queryString);
    return s.doSearch();
  }
  
  /**
   * Take a Throwable object and writes the stacktrace to a String.
   * This is used to provide that stacktrace as an insert to error
   * messages.
   * @param thr
   * @return the stacktrace.
   */
  private String stackTraceToString(Throwable thr)
  {
    StackTraceElement[] elements = thr.getStackTrace();
    StringBuffer buf = new StringBuffer();
    
    for (int i = 0; i < elements.length; i++)
    {
      StackTraceElement element = elements[i];
      buf.append(element);
      
      if(i != elements.length-1)
        buf.append(System.getProperty("line.separator"));
    }
    
    return buf.toString();
  }

  /**
   * Logs an error message to the platforms syslog/eventlog and throws a
   * MbUserException contain the information. Throwing this exception will
   * cause the broker to propagate the current message to the failure terminal
   * 
   * @param obj The source instance.
   * @param methodName The method name.
   * @param messageSrc The message source.
   * @param key The message key.
   * @param traceText The trace text, that will be written to user trace.
   * @param inserts The inserts for the message.
   * 
   * @throws MbException
   */
  private void logErrorGenerateException(Object obj,
                                         String methodName,
                                         String messageSrc,
                                         String key,
                                         String traceText, 
                                         String[] inserts) throws MbException
  {
    MbService.logError(obj,
                       methodName,
                       messageSrc,
                       key, 
                       traceText,
                       inserts);

    throw new MbUserException(obj,
                              methodName,
                              messageSrc,
                              key, 
                              traceText,
                              inserts);
  }
  /**
   * This class is the ResourceBundle containing the  messages for this example.
   */
  public static class GoogleAPINodeMesssages extends ListResourceBundle
  {
    public static final String MESSAGE_SOURCE = GoogleAPINodeMesssages.class.getName();                                                
    public static final String GOOGLE_API_EXCEPTION = "GOOGLE_API_EXCEPTION";
   
    private Object[][] messages  = {{GOOGLE_API_EXCEPTION, "An exception occurred calling the Google API: key: %1 query: %2 exception %3 stacktrace: %4" }};

    /* (non-Javadoc)
     * @see java.util.ListResourceBundle#getContents()
     */
    public Object[][] getContents()
    {
      return messages;
    }
  }
  
}

