/* 
 * Sample program for use with Product          
 *  ProgIds: 5724-J06 5724-J05 5724-J04 5697-J09 5655-M74 5655-M75 5648-C63 
 *  (C) Copyright IBM Corporation 2005.                      
 * All Rights Reserved * Licensed Materials - Property of IBM 
 * 
 * This sample program is provided AS IS and may be used, executed, 
 * copied and modified without royalty payment by customer 
 * 
 * (a) for its own instruction and study, 
 * (b) in order to develop applications designed to run with an IBM 
 *     WebSphere product, either for customer's own internal use or for 
 *     redistribution by customer, as part of such an application, in 
 *     customer's own products. 
 */

package com.ibm.broker.javacompute.samples;
import java.io.IOException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.mail.*;
import javax.mail.search.SearchTerm;

import com.ibm.broker.javacompute.MbJavaComputeNode;
import com.ibm.broker.plugin.*;

/**
 *  
 * The NewsGroupGetNode sample demonstrates how a Java compute node can call out to
 * an external API and augment this incoming message with the results of this call.
 * 
 * <p>
 * The node receives a incoming message of the following form:
 * 
 * <p>
 * <newsgroupgetnode>
 *   <newsgroup>ibm.software.websphere.mqintegrator</newsgroup> <!-- A newsgroup found news.software.ibm.com -->
 *   <searchexpression>.*Java.*</searchexpression> <!-- A regular expression search term -->
 * </newsgroupgetnode>
 * 
 * <p>
 * 
 * The "newsgroup" and "searchexpression" element values are extracted from the incoming message. 
 * The JavaMail API is then used to search the newsgroup for posting that have a subject that match 
 * the searchexpression.
 * 
 * <p>
 * 
 * The results of this search are used to augment the incoming message to form a message of
 * following format.
 * 
 * 
 * <newsgroupgetnode>
 *   <newsgroup>ibm.software.websphere.mqintegrator</newsgroup>
 *   <searchexpression> .*Java.*</searchexpression>
 *   <result>
 *    <message>
 *      <subject> ... </subject>
 *      <from> .. </from>
 *      <sentdate> ... </senddate>
 *      <body> ... </body>   
 *    </message>
 *    ...
 *   </result>
 * </newsgroupgetnode>
 * 
 */
public class NewsGroupGetNode extends MbJavaComputeNode
{
  private static final String NEWS_SOFTWARE_IBM_COM = "news.software.ibm.com";
  
  // The newsgroup server being searched.
  private static Store store;
  
  public void evaluate(MbMessageAssembly incomingAssembly) throws MbException
  {
    final String methodName = "evaluate";
    // First use XPath an expression to extract the "newsgroup" and "searchexpression" element values.
    String newsGroup = (String)incomingAssembly.getMessage().evaluateXPath("string(/newsgroupgetnode/newsgroup)");
    String searchExpression = (String)incomingAssembly.getMessage().evaluateXPath("string(/newsgroupgetnode/searchexpression)");

    try
    {
      // Search the newsgroup for the messages.
      Message[] newsMessages = findMessages(newsGroup, searchExpression);
      // Build the result message
      MbMessage resultMessage = buildResultMessage(incomingAssembly.getMessage(), newsMessages);
      getOutputTerminal("out").propagate(new MbMessageAssembly(incomingAssembly, resultMessage));
    }
    catch (MessagingException mex)
    {
      // Exception from JavaMail API, log error and throw MbUserException
      String stackTrace = stackTraceToString(mex);
      logErrorGenerateException(this, 
                                methodName,
                                NewGroupGetNodeMessage.MESSAGE_SOURCE,
                                NewGroupGetNodeMessage.JAVAMAIL_API_EXCEPTION,
                                "Exception in JavaMail API",
                                new String[]{mex.toString(), stackTrace});
    }
    catch(IOException ioex)
    {
      // IOException, log error and throw MbUserException
      String stackTrace = stackTraceToString(ioex);
      logErrorGenerateException(this, 
                                methodName,
                                NewGroupGetNodeMessage.MESSAGE_SOURCE,
                                NewGroupGetNodeMessage.IOEXCEPTION,
                                "IOException",
                                new String[]{ioex.toString(), stackTrace});
    }
  }

  /**
   * Builds the outgoing result message based on the incoming message and the result
   * of the newsgroup search.
   * 
   * @param incomingMessage The incoming message.
   * @param newsMessages The messages from the newsgroups matching the search expression
   * @return The outgoing results message.
   * @throws MbException
   * @throws MessagingException
   * @throws IOException
   */
  private MbMessage buildResultMessage(MbMessage incomingMessage, Message[] newsMessages) throws MbException, IOException, MessagingException
  {
    // Copy the incoming message to a new results message.
    MbMessage resultMessage = new MbMessage(incomingMessage);
    // Use XPath to navigate down to the "newsgroupgetnode"
    List elements = (List)resultMessage.evaluateXPath("/newsgroupgetnode");
    MbElement newsGroupGetNodeElement = (MbElement)elements.get(0);
    // Create the "result" element 
    MbElement resultElement = newsGroupGetNodeElement.createElementAsLastChild(MbElement.TYPE_NAME, "result", null );
    
    // Iterate of the newsgroup messages
    for (int i = 0; i < newsMessages.length; i++)
    {
      Message message = newsMessages[i];
      
      // Only add elements for messages that are plain text.
      if(message.getContentType().equals("text/plain"))
      {
        MbElement messageElement = resultElement.createElementAsLastChild(MbElement.TYPE_NAME, "message", null );
        messageElement.createElementAsLastChild(MbElement.TYPE_NAME, "subject", message.getSubject());
        
        Address[] fromAddresses = message.getFrom();
        if(fromAddresses.length > 0)
        {
          messageElement.createElementAsLastChild(MbElement.TYPE_NAME, 
                                                  "from", 
                                                  fromAddresses[0] == null ? null : fromAddresses[0].toString() );
        }
        messageElement.createElementAsLastChild(MbElement.TYPE_NAME, "senddate", message.getSentDate().toString());
        messageElement.createElementAsLastChild(MbElement.TYPE_NAME, "body", (String)message.getContent());
      }
    }
    
    return resultMessage;
  }
  
  /**
   * Returns the message store for the news.software.ibm.com news server.
   * 
   * @return The message store.
   * @throws MessagingException
   */
  private static Store getStore() throws MessagingException
  {
    // Only create if first call or not connected
    if(store == null || !store.isConnected())
    {
      Properties props = new Properties();

      // Get session
      Session session = Session.getInstance( props, null);

      // Get the store
      store = session.getStore("nntp");
      store.connect(NEWS_SOFTWARE_IBM_COM, null, null);
    }
    
    return store;
  }
  
  /**
   * Search the newsgroup for messages match the searchExpression. Returns an array of matching 
   * messages.
   * 
   * @param newsGroup The newsgroup to search.
   * @param searchExpression The search regular expression.
   * @return The match messages.
   * @throws MessagingException
   */
  private static Message[] findMessages(String newsGroup, String searchExpression) throws MessagingException
  {
    Folder folder = getStore().getFolder(newsGroup);
    folder.open(Folder.READ_ONLY);
    
    Message[] messages = folder.search(new RegexSearchTerm(searchExpression));
    
    return messages;
  }
  
  /**
   * Provides an implementation of SearchTerm using regular expression.
   * 
   */
  private static class RegexSearchTerm extends SearchTerm
  {
    Pattern regex;
    private RegexSearchTerm(String regex)
    {
      this.regex = Pattern.compile(regex);
    }
    
    /* (non-Javadoc)
     * @see javax.mail.search.SearchTerm#match(javax.mail.Message)
     */
    public boolean match(Message msg)
    {
      boolean matches = false;
      
      Matcher m;
      try
      {
        m = regex.matcher(msg.getSubject());
        matches = m.matches();
      }
      catch (MessagingException e)
      {
        // If an messaging exception occurs then return false;
      }
      
      return matches;
    }
  }
  
  /**
   * Takes a Throwable object and writes the stacktrace to a String.
   * This is used to provide that stacktrace as an insert to error
   * messages.
   * @param thr
   * @return the stacktrace.
   */
  private String stackTraceToString(Throwable thr)
  {
    StackTraceElement[] elements = thr.getStackTrace();
    StringBuffer buf = new StringBuffer();
    
    for (int i = 0; i < elements.length; i++)
    {
      StackTraceElement element = elements[i];
      buf.append(element);
      
      if(i != elements.length-1)
        buf.append(System.getProperty("line.separator"));
    }
    
    return buf.toString();
  }

  /**
   * Logs an error message to the platforms syslog/eventlog and throws a
   * MbUserException contain the information. Throwing this exception will
   * cause the broker to propagate the current message to the failure terminal
   * 
   * @param obj The source instance.
   * @param methodName The method name.
   * @param messageSrc The message source.
   * @param key The message key.
   * @param traceText The trace text, that will be written to user trace.
   * @param inserts The inserts for the message.
   * 
   * @throws MbException
   */
  private void logErrorGenerateException(Object obj,
                                         String methodName,
                                         String messageSrc,
                                         String key,
                                         String traceText, 
                                         String[] inserts) throws MbException
  {
    MbService.logError(obj,
                       methodName,
                       messageSrc,
                       key, 
                       traceText,
                       inserts);

    throw new MbUserException(obj,
                              methodName,
                              messageSrc,
                              key, 
                              traceText,
                              inserts);
  }
  /**
   * The class is the ResourceBundle containg all the messages for this example.
   */
  public static class NewGroupGetNodeMessage extends ListResourceBundle
  {
    public static final String MESSAGE_SOURCE = NewGroupGetNodeMessage.class.getName();                                                
    public static final String JAVAMAIL_API_EXCEPTION = "JAVAMAIL_API_EXCEPTION";
    public static final String IOEXCEPTION = "IOEXCEPTION";
   
    private Object[][] messages  = {{JAVAMAIL_API_EXCEPTION, "An exception occurred calling the JavaMail API: exception %1 stacktrace: %2" },
                                    {IOEXCEPTION, "An IOException has been thrown: exception %1 stacktrace: %2" }};
                                    

    /* (non-Javadoc)
     * @see java.util.ListResourceBundle#getContents()
     */
    public Object[][] getContents()
    {
      return messages;
    }
  }
  
}



